void add_script_to_cache(systemtap_session& s);
bool get_script_from_cache(systemtap_session& s);

void add_stapconf_to_cache(systemtap_session& s);
bool get_stapconf_from_cache(systemtap_session& s);

void clean_cache(systemtap_session& s);

/* vim: set sw=2 ts=8 cino=>4,n-2,{2,^-2,t0,(0,u0,w1,M1 : */
