/* Systemtap Debug Macros
 * Copyright (C) 2012 Red Hat Inc.
 *
 * This file is part of systemtap, and is free software.  You can
 * redistribute it and/or modify it under the terms of the GNU General
 * Public License (GPL); either version 2, or (at your option) any
 * later version.
 */

#ifndef _STAPDYN_DEBUG_H_
#define _STAPDYN_DEBUG_H_

// TODO fill in this file as needed!

#endif /* _STAPDYN_DEBUG_H_ */
